#!/usr/bin/env python

"""Tests for `xlavir.tools.nextflow` package."""

from pathlib import Path

from xlavir.tools import fastp

dirpath = Path(__file__).parent

expected = {
    'Sample1': 133769420,
    'Sample2': 523300,
}


def test_fastp_get_total_reads():
    """Test finding and parsing of Nextflow workflow execution HTML report"""
    sample_reads = fastp.get_info(dirpath)
    assert sample_reads == expected



from pydantic import BaseModel


class QualityRequirements(BaseModel):
    min_genome_coverage: float = 0.95
    min_median_depth: int = 30
    low_coverage_threshold: int = 10
    major_allele_freq: float = 0.75



"""VCF and SnpEff/SnpSift parsing functions"""
import logging
import os
import re
from dataclasses import dataclass
from operator import itemgetter
from pathlib import Path
from typing import Dict, Tuple, List, Optional, Iterable, Union

import pandas as pd
from pydantic import BaseModel

from xlavir.qc import QualityRequirements
from xlavir.util import try_parse_number, find_file_for_each_sample

logger = logging.getLogger(__name__)

aa_codes = dict(
    ALA='A',
    ARG='R',
    ASN='N',
    ASP='D',
    CYS='C',
    GLU='E',
    GLN='Q',
    GLY='G',
    HIS='H',
    ILE='I',
    LEU='L',
    LYS='K',
    MET='M',
    PHE='F',
    PRO='P',
    SER='S',
    THR='T',
    TRP='W',
    TYR='Y',
    VAL='V',
    TER='*',
)

# list of tuples: [0]: column id; [1]: report column id/name; [2]: column description for report cell comment
variants_cols = [
    ('sample', 'Sample', 'Sample name',),
    ('CHROM', 'Reference Genome', 'Reference genome sequence ID/name'),
    (
        'mutation',
        'Mutation',
        'Mutation found in sample with format '
        '"{reference allele}{reference position}{allele in sample}"'
        ' with predicted amino acid change information in brackets with format '
        '"{gene name}:{reference AA}{gene AA position}{AA change}"'
    ),
    ('POS', 'Position', '1-based nucleotide position in reference sequence'),
    ('REF', 'Reference Allele', 'Nucleotide allele sequence found in reference sequence'),
    ('ALT', 'Alternate Allele', 'Nucleotide allele sequence found in sample'),
    (
        'REF_DP',
        'Reference Allele Depth',
        'Read depth of coverage supporting reference allele at reference position'
    ),
    (
        'ALT_DP',
        'Alternate Allele Depth',
        'Read depth of coverage supporting alternate allele at reference position',
    ),
    ('DP', 'Total Depth', 'Total depth of coverage at reference position',),
    (
        'ALT_FREQ',
        'Alternate Allele Frequency',
        'Observed frequency of alternate allele variant',
    ),
    ('gene', 'Gene', 'Gene name',),
    (
        'impact',
        'Variant Impact',
        'SnpEff estimation of putative impact or deleteriousness of variant '
        '(see https://pcingola.github.io/SnpEff/se_inputoutput/#ann-field-vcf-output-files)'
    ),
    (
        'effect',
        'Variant Effect',
        'Effect of variant annotated using Sequence Ontology terms, e.g.'
        'for "missense_variant", see http://www.sequenceontology.org/browser/current_release/term/SO:0001583'
        ' where the definition is "A sequence variant, that changes one or more bases, resulting in a '
        'different amino acid sequence but where the length is preserved."'
    ),
    ('aa', 'Amino Acid Change', 'The change in the sample\'s gene amino acid sequence'
                                ' relative to the reference sequence'),
    ('aa_pos', 'Amino Acid Position', 'Position of amino acid change in the reference sequence gene'),
    ('aa_len', 'Gene Amino Acid Length', 'Amino acid length of the reference sequence gene'),
]

variant_summary_cols = [
    (
        'Mutation',
        'Mutation',
        'Mutation found in sample with format '
        '"{reference allele}{reference position}{allele in sample}"'
        ' with predicted amino acid change information in brackets with format '
        '"{gene name}:{reference AA}{gene AA position}{AA change}"'
    ),
    ('n_samples', '# of Samples', 'Number of samples with the mutation.'),
    ('samples', 'Samples', 'List of samples with mutation delimited by semicolon (";")'),
    (
        'min_depth',
        'Min Depth',
        'Minimum depth in all samples that the mutation is observed at.'
    ),
    (
        'max_depth',
        'Max Depth',
        'Maximum depth in all samples that the mutation is observed at.'
    ),
    (
        'mean_depth',
        'Mean Depth',
        'Mean/average depth that the mutation is observed at.'
    ),
    (
        'min_af',
        'Min AF',
        'Minimum alternate allele frequency of mutation in all samples.'
    ),
    (
        'max_af',
        'Max AF',
        'Maximum alternate allele frequency of mutation in all samples.'
    ),
    (
        'mean_af',
        'Mean AF',
        'Mean/average alternate allele frequency of mutation in all samples.'
    ),
    (
        'nt_pos',
        'Nucleotide Position',
        'Nucleotide position of mutation with respect to reference genome.'
    ),
    (
        'aa_pos',
        'Amino Acid Position',
        'Amino acid position of mutation in reference genome gene.'
    )
]

BCFTOOLS_STATS_GLOB_PATTERNS = [
    '**/ivar/**/*AF0.*.bcftools_stats.txt',
    '**/ivar/**/*.bcftools_stats.txt',
    '**/*AF0.*.bcftools_stats.txt',
    '**/*.bcftools_stats.txt',
]

BCFTOOLS_STATS_SAMPLE_NAME_CLEANUP = [
    re.compile(r'\.AF0\.\d+\.bcftools_stats.txt$')
]


class VariantStats(BaseModel):
    sample: str
    n_snp: int
    n_mnp: int
    n_indel: int


@dataclass
class VariantCaller:
    iVar = 'iVar'
    Longshot = 'Longshot'
    Nanopolish = 'nanopolish'
    Medaka = 'medaka'
    Clair3 = 'clair3'
    Bcftools = 'bcftools'


VCF_GLOB_PATTERNS = [
    '**/nanopolish/*.pass.vcf.gz',
    '**/ivar/*.vcf.gz',
    '**/bcftools/*.vcf.gz',
    '**/*.filt.no_fs.vcf',
    '**/*.longshot.vcf',
    '**/*.vcf',
    '**/*.vcf.gz',
]

VCF_SAMPLE_NAME_CLEANUP = [
    re.compile(r'(\.pass)?\.vcf(\.gz)?$'),
    re.compile(r'\.AF0\.\d+(\.filt)?'),
    re.compile(r'\.0\.\d+AF(\.filt)?'),
    re.compile(r'\.medaka'),
    re.compile(r'\.longshot'),
    re.compile(r'\.clair3'),
    re.compile(r'\.snpeff'),
    re.compile(r'\.no_fs'),
    re.compile(r'\.merged'),
    re.compile(r'\.filtered'),
]

SNPSIFT_GLOB_PATTERNS = [
    '**/ivar/**/*.snpSift.table.txt',
    '**/ivar/**/*.snpsift.table.txt',
    '**/ivar/**/*.snpsift.txt',
    '**/*.snpSift.table.txt',
    '**/*.snpsift.table.txt',
    '**/*.snpsift.txt',
]

SNPSIFT_SAMPLE_NAME_CLEANUP = [
    re.compile(r'\.snp[sS]ift\.table\.txt$'),
    re.compile(r'\.snp[sS]ift\.txt$'),
    re.compile(r'\.AF0\.\d+(\.filt)?'),
    re.compile(r'\.0\.\d+AF(\.filt)?'),
    re.compile(r'\.filtered'),
]


def vcf_selector(paths: List[Path]) -> Optional[Path]:
    xs = []
    for path in paths:
        variant_caller, df = read_vcf(path)
        xs.append((df.shape[0], path))
    xs.sort(reverse=True)
    try:
        return xs[0][1]
    except KeyError:
        return None


def snpsift_selector(paths: List[Path]) -> Optional[Path]:
    xs = []
    for path in paths:
        df = pd.read_table(path)
        xs.append((df.shape[0], path))
    xs.sort(reverse=True)
    try:
        return xs[0][1]
    except KeyError:
        return None


def read_vcf(vcf_file: Path) -> Tuple[str, pd.DataFrame]:
    """Read VCF file into a DataFrame"""
    gzipped = vcf_file.name.endswith('.gz')
    with os.popen(f'zcat < {vcf_file.absolute()}') if gzipped else open(vcf_file) as fh:
        vcf_cols = []
        variant_caller = ''
        for line in fh:
            if line.startswith('##source='):
                variant_caller = line.strip().replace('##source=', '')
            if line.startswith('##bcftools_callVersion'):
                variant_caller = 'bcftools'
            if line.startswith('##nanopolish'):
                variant_caller = 'nanopolish'
            if line.startswith('##medaka_version'):
                variant_caller = 'medaka'
            if line.startswith('#CHROM'):
                vcf_cols = line[1:].strip().split('\t')
                break
        try:

            df = pd.read_table(vcf_file,
                               sep='\t',
                               comment='#',
                               header=None,
                               names=vcf_cols)
            df = df[~df.duplicated(['CHROM', 'POS', 'ID', 'REF', 'ALT', 'FILTER'], keep='first')]
        except pd.errors.EmptyDataError:
            return variant_caller, pd.DataFrame()
    return variant_caller, df


def parse_aa(gene: str,
             ref: str,
             alt: str,
             nt_pos: int,
             aa_pos: int,
             snpeff_aa: str,
             effect: str) -> str:
    if snpeff_aa == '.':
        return f'{ref}{nt_pos}{alt}'
    m = re.match(r'p\.([a-zA-Z]+)(\d+)([a-zA-Z]+)', snpeff_aa)
    if m is None and snpeff_aa.startswith('p.'):
        aa_str = snpeff_aa[2:].upper()
        for aa3, aa1 in aa_codes.items():
            aa_str = aa_str.replace(aa3, aa1)
        if aa_str.endswith('DEL'):
            aa_str = aa_str.replace('DEL', 'del')
        return f'{gene}:{aa_str} ({ref}{nt_pos}{alt})'
    ref_aa, aa_pos_str, alt_aa = m.groups()
    ref_aa = get_aa(ref_aa)

    if effect == 'stop_lost':
        alt_aa = get_aa(alt_aa.replace('ext', ''))
        return f'{gene}:{ref_aa}{aa_pos_str}{alt_aa} ({ref}{nt_pos}{alt} [stop_lost])'
    if effect == 'frameshift_variant':
        return f'{gene}:{ref_aa}{aa_pos_str} ({ref}{nt_pos}{alt} [FRAMESHIFT])'
    if effect == 'conservative_inframe_deletion':
        return f'{gene}:{ref_aa}{aa_pos_str}{alt_aa} ({ref}{nt_pos}{alt})'
    if effect == 'disruptive_inframe_deletion':
        return f'{gene}:{ref_aa}{aa_pos_str}{alt_aa} ({ref}{nt_pos}{alt} [disruptive_inframe_deletion])'

    alt_aa = get_aa(alt_aa)
    if alt_aa == ref_aa:
        return f'{ref}{nt_pos}{alt}'
    return f'{gene}:{ref_aa}{aa_pos_str}{alt_aa} ({ref}{nt_pos}{alt})'


def get_aa(s: str) -> str:
    out = ''
    for i in range(0, len(s), 3):
        aa = s[i: i + 3]
        try:
            aa_code = aa_codes[aa.upper()]
        except KeyError:
            aa_code = aa
        out += aa_code
    return out


def simplify_snpsift(df: pd.DataFrame, sample_name: str) -> Optional[pd.DataFrame]:
    if df.empty:
        return None
    df = df[~df.duplicated(keep='first')]
    field_names = set()
    series = []
    for c in df.columns:
        if c == 'AC':
            df_ac = df[c].str.split(',', n=1, expand=True)
            REF_AC = df_ac[0].astype(int)
            REF_AC.name = 'REF_AC'
            ALT_AC = df_ac[1].astype(int)
            ALT_AC.name = 'ALT_AC'
            AF = ALT_AC / (REF_AC + ALT_AC)
            AF.name = 'AF'
            series += [REF_AC, ALT_AC, AF]
            continue
        if c == 'SR':
            df_ac = df[c].str.split(',', expand=True)
            REF_AC = df_ac[0].astype(int) + df_ac[1].astype(int)
            REF_AC.name = 'REF_AC'
            ALT_AC = df_ac[2].astype(int) + df_ac[3].astype(int)
            ALT_AC.name = 'ALT_AC'
            AF = ALT_AC / (REF_AC + ALT_AC)
            AF.name = 'AF'
            series += [REF_AC, ALT_AC, AF]
            continue
        idx = c.find('[*].')
        if idx > 0:
            new_series_name = c[idx + 4:].lower()
            if new_series_name in field_names:
                continue
            else:
                field_names.add(new_series_name)
            dfc = df[c]
            if dfc.dtype == 'object' and isinstance(dfc.values[0], str):
                new_series = dfc.str.split(',', n=1, expand=True)[0]
            else:
                new_series = dfc.astype('str')
            new_series.name = new_series_name
            series.append(new_series)
        else:
            series.append(df[c])
    df_out = pd.concat(series, axis=1)
    mutation_desc = [
        parse_aa(
            gene=row.gene,
            ref=row.REF,
            alt=row.ALT,
            nt_pos=row.POS,
            aa_pos=row.aa_pos,
            snpeff_aa=row.aa,
            effect=row.effect,
        )
        for row in df_out.itertuples()
    ]
    df_out['mutation'] = mutation_desc
    df_out['sample'] = sample_name
    return df_out


def parse_ivar_vcf(df: pd.DataFrame, sample_name: str = None) -> Optional[pd.DataFrame]:
    if df.empty:
        return None
    if not sample_name:
        sample_name = df.columns[-1] if df.columns[-1] != 'SAMPLE' else None
        if sample_name is None:
            raise ValueError(f'Sample name is not defined for VCF: shape={df.shape}; columns={df.columns}')
    pos_fmt_val = {}
    for row in df.itertuples():
        # An iVar VCF has the DP, total depth, in the INFO field, you cannot sum the alt and ref depths since the ref
        # depth only applies to the first base in the ref allele rather than over the entire allele so the depth can be
        # misleading for deletions
        # See iVar issue: https://github.com/andersen-lab/ivar/issues/86
        infos = parse_vcf_info(row.INFO)
        total_dp = infos['DP']
        ks = row.FORMAT.split(':')
        vs = row[-1].split(':')
        record: Dict[str, Union[float, int, str]] = {k: try_parse_number(v) for k, v in zip(ks, vs)}
        ref_dp = record['REF_DP']
        alt_dp = record['ALT_DP']
        # if the sum of the ref and alt dp does not equal the total dp reported by iVar then recalculate the ref dp
        # since it is likely only reporting the ref dp for the first base of a longer deletion. SNPs should be fine.
        sum_ref_alt_dp = ref_dp + alt_dp
        if sum_ref_alt_dp != total_dp:
            record['REF_DP'] = total_dp - alt_dp
            logger.warning(
                f'iVar VCF for sample "{sample_name}" contains a variant at position {row.POS} where sum of ref allele '
                f'depth ({ref_dp}) and alt allele depth ({alt_dp}) does not equal the total depth ({total_dp}) '
                f'reported by iVar (i.e. {ref_dp} + {alt_dp} = {sum_ref_alt_dp}; {sum_ref_alt_dp} != {total_dp}). Ref '
                f'allele is "{row.REF}", alt allele is "{row.ALT}". iVar alt allele frequency: {record["ALT_FREQ"]}')
        pos_fmt_val[row.POS] = record
        pos_fmt_val[row.POS]['DP'] = total_dp
    df_ivar_info = pd.DataFrame(pos_fmt_val).transpose()
    df_ivar_info.index.name = 'POS'
    df_ivar_info.reset_index(inplace=True)
    df_merge = pd.merge(df, df_ivar_info, on='POS')
    df_merge['sample'] = sample_name
    return df_merge.drop(columns=['ID', 'INFO', 'QUAL', 'FILTER', 'FORMAT', df.columns[-1], 'GT'])


def merge_vcf_snpsift(df_vcf: Optional[pd.DataFrame],
                      df_snpsift: Optional[pd.DataFrame]) -> Optional[pd.DataFrame]:
    if df_snpsift is None and df_vcf is None:
        return None
    if df_vcf is None:
        snpsift_cols = set(df_snpsift.columns)
        return df_snpsift.loc[:, [x for x, _, _ in variants_cols if x in snpsift_cols]]
    if df_snpsift is None:
        vcf_cols = set(df_vcf.columns)
        return df_vcf.loc[:, [x for x, _, _ in variants_cols if x in vcf_cols]]

    df_merge = pd.merge(df_vcf, df_snpsift)
    merged_cols = set(df_merge.columns)
    return df_merge.loc[:, [x for x, _, _ in variants_cols if x in merged_cols]]


def parse_longshot_vcf(df: pd.DataFrame, sample_name: str = None) -> Optional[pd.DataFrame]:
    if df.empty:
        return None
    if not sample_name:
        sample_name = df.columns[-1] if df.columns[-1] != 'SAMPLE' else None
        if sample_name is None:
            raise ValueError(f'Sample name is not defined for VCF: shape={df.shape}; columns={df.columns}')
    pos_info_val = {}
    for row in df.itertuples():
        infos = parse_vcf_info(row.INFO)
        ac_ref, ac_alt = infos['AC']
        infos['REF_DP'] = ac_ref
        infos['ALT_DP'] = ac_alt
        pos_info_val[row.POS] = infos
    df_longshot_info = pd.DataFrame(pos_info_val).transpose()
    df_longshot_info.index.name = 'POS'
    df_longshot_info.reset_index(inplace=True)
    df_merge = pd.merge(df, df_longshot_info, on='POS')
    df_merge['sample'] = sample_name
    df_merge = df_merge[df_merge.DP > 0]
    df_merge['ALT_FREQ'] = df_merge.ALT_DP / df_merge.DP
    cols_to_keep = list({col for col, _, _ in variants_cols} & set(df_merge.columns))
    return df_merge.loc[:, cols_to_keep]


def parse_medaka_vcf(
        df: pd.DataFrame,
        sample_name: str = None,
        qc_reqs: QualityRequirements = None
) -> Optional[pd.DataFrame]:
    if df.empty:
        return None
    if not sample_name:
        sample_name = df.columns[-1] if df.columns[-1] != 'SAMPLE' else None
        if sample_name is None:
            raise ValueError(f'Sample name is not defined for VCF: shape={df.shape}; columns={df.columns}')
    pos_info_val = {}
    for row in df.itertuples():
        # if there is no variant INFO available then this isn't the right file
        if row.INFO == '.':
            return None
        infos = parse_vcf_info(row.INFO)
        # no DP INFO? skip this file
        if 'DP' not in infos:
            return None
        if infos['DP'] < qc_reqs.low_coverage_threshold:
            continue
        ac_ref_fwd, ac_ref_rev, ac_alt_fwd, ac_alt_rev = infos['SR']
        infos['REF_DP'] = ac_ref_fwd + ac_ref_rev
        infos['ALT_DP'] = ac_alt_fwd + ac_alt_rev
        if infos['REF_DP'] + infos['ALT_DP'] < qc_reqs.low_coverage_threshold:
            continue
        pos_info_val[row.POS] = infos
    if pos_info_val == {}:
        return None
    df_medaka_info = pd.DataFrame(pos_info_val).transpose()
    df_medaka_info.index.name = 'POS'
    df_medaka_info.reset_index(inplace=True)
    df_merge = pd.merge(df, df_medaka_info, on='POS')
    df_merge['sample'] = sample_name
    df_merge = df_merge[df_merge.DP > 0]
    df_merge['ALT_FREQ'] = df_merge.ALT_DP / (df_merge.ALT_DP + df_merge.REF_DP)
    cols_to_keep = list({col for col, _, _ in variants_cols} & set(df_merge.columns))
    return df_merge.loc[:, cols_to_keep]


def parse_nanopolish_vcf(
        df: pd.DataFrame,
        sample_name: str = None
) -> Optional[pd.DataFrame]:
    if df.empty:
        return None
    if not sample_name:
        sample_name = df.columns[-1] if df.columns[-1] != 'sample' else None
        if sample_name is None:
            raise ValueError(
                f'Sample name is not defined for VCF: shape={df.shape}; '
                f'columns={df.columns}'
            )
    pos_info_val = {}
    for row in df.itertuples():
        infos = parse_vcf_info(row.INFO)
        fwd_ref, rev_ref, fwd_alt, rev_alt = infos['StrandSupport']
        allele_count = infos['AlleleCount']
        if allele_count > 1:
            raise NotImplementedError(
                f'Handling of allele count of {allele_count} is not supported. '
                f'Only allele counts of 1 are supported.'
            )
        infos['DP'] = fwd_ref + fwd_alt + rev_ref + rev_alt
        infos['REF_DP'] = fwd_ref + rev_ref
        infos['ALT_DP'] = fwd_alt + rev_alt
        pos_info_val[row.POS] = infos
    df_nanopolish_info = pd.DataFrame(pos_info_val).transpose()
    df_nanopolish_info.index.name = 'POS'
    df_nanopolish_info.reset_index(inplace=True)
    df_merge = pd.merge(df, df_nanopolish_info, on='POS')
    df_merge['sample'] = sample_name
    df_merge = df_merge[df_merge.DP > 0]
    df_merge['ALT_FREQ'] = df_merge.ALT_DP / df_merge.DP
    cols_to_keep = list({col for col, _, _ in variants_cols} & set(df_merge.columns))
    return df_merge.loc[:, cols_to_keep]

def parse_bcftools_vcf(df: pd.DataFrame, sample_name: str = None) -> Optional[pd.DataFrame]:
    if df.empty:
        return None
    if not sample_name:
        sample_name = df.columns[-1] if df.columns[-1] != 'sample' else None
        if sample_name is None:
            raise ValueError(f'Sample name is not defined for VCF: shape={df.shape}; columns={df.columns}')
    pos_info_val = {}
    for row in df.itertuples():
        infos = parse_vcf_info(row.INFO)
        allele_count = infos['AC']
        if allele_count > 1:
            raise NotImplementedError(f'Handling of allele count of {allele_count} is not supported. '
                                      f'Only allele counts of 1 are supported.')
        ref_dp, alt_dp = infos['AD']
        infos['REF_DP'] = ref_dp
        infos['ALT_DP'] = alt_dp
        pos_info_val[row.POS] = infos
    df_bcftools_info = pd.DataFrame(pos_info_val).transpose()
    df_bcftools_info.index.name = 'POS'
    df_bcftools_info.reset_index(inplace=True)
    df_merge = pd.merge(df, df_bcftools_info, on='POS')
    df_merge['sample'] = sample_name
    df_merge = df_merge[df_merge.DP > 0]
    df_merge['ALT_FREQ'] = df_merge.ALT_DP / df_merge.DP
    cols_to_keep = list({col for col, _, _ in variants_cols} & set(df_merge.columns))
    return df_merge.loc[:, cols_to_keep]


def parse_vcf_info(s: str) -> dict:
    out = {}
    for x in s.split(';'):
        if not x:
            continue
        if '=' not in x:
            continue
        key, val_str = x.split('=', maxsplit=1)
        out[key] = try_parse_number(val_str)
    return out


def parse_clair3_vcf(
        df: pd.DataFrame,
        sample_name: str,
        qc_reqs: QualityRequirements
) -> Optional[pd.DataFrame]:
    if df.empty:
        return None
    if not sample_name:
        sample_name = df.columns[-1] if df.columns[-1] != 'SAMPLE' else None
        if sample_name is None:
            raise ValueError(f'Sample name is not defined for VCF: shape={df.shape}; columns={df.columns}')
    pos_info_val = {}
    for row in df.itertuples():
        infos = parse_vcf_info(row.INFO)
        # no DP INFO? skip this file
        if 'DP' not in infos:
            return None
        depth = infos['DP']
        if depth < qc_reqs.low_coverage_threshold:
            continue
        allele_fraction = infos['AF']
        alt_dp = int(depth * allele_fraction)
        ref_dp = depth - alt_dp
        infos['ALT_DP'] = alt_dp
        infos['REF_DP'] = ref_dp
        if infos['REF_DP'] + infos['ALT_DP'] < qc_reqs.low_coverage_threshold:
            continue
        pos_info_val[row.POS] = infos
    if pos_info_val == {}:
        return None
    df_clair3_info = pd.DataFrame(pos_info_val).transpose()
    df_clair3_info.index.name = 'POS'
    df_clair3_info.reset_index(inplace=True)
    df_merge = pd.merge(df, df_clair3_info, on='POS')
    df_merge['sample'] = sample_name
    df_merge = df_merge[df_merge.DP > 0]
    df_merge['ALT_FREQ'] = df_merge.AF
    cols_to_keep = list({col for col, _, _ in variants_cols} & set(df_merge.columns))
    return df_merge.loc[:, cols_to_keep]


def get_info(basedir: Path, qc_reqs: QualityRequirements) -> Dict[str, pd.DataFrame]:
    sample_vcf = find_file_for_each_sample(basedir=basedir,
                                           glob_patterns=VCF_GLOB_PATTERNS,
                                           sample_name_cleanup=VCF_SAMPLE_NAME_CLEANUP,
                                           single_entry_selector_func=vcf_selector)
    sample_dfvcf = {}
    for sample, vcf_path in sample_vcf.items():
        variant_caller, df_vcf = read_vcf(vcf_path)
        if variant_caller.startswith(VariantCaller.iVar):
            df_parsed_ivar_vcf = parse_ivar_vcf(df_vcf, sample)
            if df_parsed_ivar_vcf is not None:
                sample_dfvcf[sample] = df_parsed_ivar_vcf
            else:
                logger.warning(f'Sample "{sample}" has no entries in VCF "{vcf_path}"')
        elif variant_caller.startswith(VariantCaller.Bcftools):
            df_bcftools_vcf = parse_bcftools_vcf(df_vcf, sample)
            if df_bcftools_vcf is not None:
                sample_dfvcf[sample] = df_bcftools_vcf
            else:
                logger.warning(f'Sample "{sample}" has no entries in VCF "{vcf_path}"')
        elif variant_caller.startswith(VariantCaller.Clair3):
            df_clair3_vcf = parse_clair3_vcf(df_vcf, sample, qc_reqs)
            if df_clair3_vcf is not None:
                sample_dfvcf[sample] = df_clair3_vcf
            else:
                logger.warning(f'Sample "{sample}" has no entries in Clair3 VCF "{vcf_path}"')
        elif variant_caller.startswith(VariantCaller.Medaka):
            df_medaka_vcf = parse_medaka_vcf(df_vcf, sample, qc_reqs)
            if df_medaka_vcf is not None:
                sample_dfvcf[sample] = df_medaka_vcf
            else:
                logger.warning(f'Sample "{sample}" has no entries in VCF "{vcf_path}"')
        elif variant_caller.startswith(VariantCaller.Longshot):
            df_longshot_vcf = parse_longshot_vcf(df_vcf, sample)
            if df_longshot_vcf is not None:
                sample_dfvcf[sample] = df_longshot_vcf
            else:
                logger.warning(f'Sample "{sample}" has no entries in VCF "{vcf_path}"')
        elif variant_caller.startswith(VariantCaller.Nanopolish):
            df_nanopolish_vcf = parse_nanopolish_vcf(df_vcf, sample)
            if df_nanopolish_vcf is not None:
                sample_dfvcf[sample] = df_nanopolish_vcf
            else:
                logger.warning(f'Sample "{sample}" has no entries in VCF "{vcf_path}"')
        else:
            logger.warning(
                f'Sample "{sample}" VCF file "{vcf_path}" with '
                f'variant_caller={variant_caller} not supported. Skipping...'
            )

    sample_snpsift = find_file_for_each_sample(basedir=basedir,
                                               glob_patterns=SNPSIFT_GLOB_PATTERNS,
                                               sample_name_cleanup=SNPSIFT_SAMPLE_NAME_CLEANUP,
                                               single_entry_selector_func=snpsift_selector)
    if not sample_snpsift:
        logger.warning(f'No SnpSift tables found in "{basedir}" using glob patterns "{SNPSIFT_GLOB_PATTERNS}"')
    sample_dfsnpsift = {}
    for sample, snpsift_path in sample_snpsift.items():
        df_snpsift = simplify_snpsift(pd.read_table(snpsift_path), sample)
        if df_snpsift is not None:
            sample_dfsnpsift[sample] = df_snpsift
        else:
            logger.warning(f'Sample "{sample}" has no entries in VCF "{snpsift_path}"')
    out = {}
    set_vcf_samples = set(sample_dfvcf.keys())
    set_snpsift_samples = set(sample_dfsnpsift.keys())
    all_samples = set_vcf_samples | set_snpsift_samples
    logger.debug(f'all_samples={len(all_samples)} | '
                 f'vcf only samples={set_vcf_samples - set_snpsift_samples} |'
                 f'snpsift only samples={set_snpsift_samples - set_vcf_samples}')
    for sample in all_samples:
        df_snpsift = sample_dfsnpsift.get(sample, None)
        df_vcf = sample_dfvcf.get(sample, None)
        df_merged = merge_vcf_snpsift(df_vcf, df_snpsift)
        if df_merged is None:
            continue
        out[sample] = df_merged

    return out


def to_dataframe(dfs: Iterable[pd.DataFrame]) -> pd.DataFrame:
    df = pd.concat(list(dfs))
    df.sort_values(['sample', 'POS'], inplace=True)
    df.set_index('sample', inplace=True)
    df.index.name = 'Sample'
    return df.rename(columns={x: y for x, y, _ in variants_cols})


def get_nt_position_int(s: str) -> int:
    """Get the nucleotide position from the mutation string

    >>> get_nt_position_int('A1879G')
    1879
    >>> get_nt_position_int('S:Y145_H146del (TATTACC21992T)')
    21992
    """
    if ':' in s:
        return int(re.sub(r'.*\([AGTC]+(\d+).*', r'\1', s))
    else:
        return int(re.sub(r'[AGTC]+(\d+).*', r'\1', s))


def to_variant_pivot_table(df: pd.DataFrame) -> pd.DataFrame:
    df_vars = df.copy()
    df_vars.reset_index(inplace=True)
    df_pivot = pd.pivot_table(df_vars,
                              index='Sample',
                              columns='Mutation',
                              values='Alternate Allele Frequency',
                              aggfunc='first',
                              fill_value=0.0)
    nt_positions = [get_nt_position_int(x) for x in df_pivot.columns]
    pivot_cols = list(zip(df_pivot.columns,
                          nt_positions))
    pivot_cols.sort(key=itemgetter(1))
    return df_pivot[[x for x, y in pivot_cols]]


def to_summary(df: pd.DataFrame) -> pd.DataFrame:
    df_vars = df.copy()
    df_vars.reset_index(inplace=True)
    logger.debug(f'df_vars columns: {df_vars.columns}')
    df_summary = df_vars.groupby('Mutation', sort=False).agg(
        n_samples=('Sample', 'size'),
        samples=('Sample', lambda x: '; '.join(x)),
        gene=('Gene', 'first'),
        effect=('Variant Effect', 'first'),
        impact=('Variant Impact', 'first'),
        aa=('Amino Acid Change', 'first'),
        min_depth=('Alternate Allele Depth', 'min'),
        max_depth=('Alternate Allele Depth', 'max'),
        mean_depth=('Alternate Allele Depth', lambda x: sum(x) / len(x)),
        min_af=('Alternate Allele Frequency', 'min'),
        max_af=('Alternate Allele Frequency', 'max'),
        mean_af=('Alternate Allele Frequency', lambda x: sum(x) / len(x)),
        nt_pos=('Position', 'first'),
        aa_pos=('Amino Acid Position', 'first')
    )
    df_summary.sort_values('nt_pos', inplace=True)
    return df_summary.rename(columns={x: y for x, y, _ in (variant_summary_cols + variants_cols)})



import re
from pathlib import Path
from typing import Optional
import logging

import pandas as pd

from xlavir.util import find_file_for_each_sample

logger = logging.getLogger(__name__)

PANGOLIN_CSV = 'pangolin.csv'

PANGOLIN_GLOB_PATTERNS = [
    '**/*.pangolin.csv',
    '**/pangolin.csv',
]

PANGOLIN_SAMPLE_NAME_CLEANUP = [
    re.compile(r'\.pangolin\.csv$'),
]

pangolin_cols = [
    ('taxon', 'Sample', 'Sample name'),
    (
        'lineage',
        'Pangolin Lineage',
        'Pangolin global SARS-CoV-2 lineage assignment based on pangoLEARN model. '
        'For more info, see https://github.com/cov-lineages/pangolin/#pangolearn-description',
    ),
    (
        'probability',
        'Lineage Assignment Probability',
        'Pangolin lineage assignment probability from multinomial logistic regression. For more info, '
        'see https://github.com/cov-lineages/pangolin/#pangolearn-description'
    ),
    (
        'conflict',
        'Conflict',
        'This is a measure of conflicts within the decision tree, 0 being '
        'equivalent to no conflicts, but not a confidence score.'
    ),
    (
        'pangolin_version',
        'Pangolin Version',
        'Version of Pangolin (https://github.com/cov-lineages/pangolin) used for lineage assignment.'
    ),
    (
        'pango_version',
        'Pango Version',
        'pangoLEARN PANGO_VERSION (https://github.com/cov-lineages/pangoLEARN/)'
    ),
    (
        'pangoLEARN_version',
        'pangoLEARN Lineages Version',
        'Release version of pangoLEARN SARS-CoV-2 '
        'lineages information used for assignment.'),
    (
        'status',
        'Pangolin QC Status',
        'QC status of Pangolin lineage assignment, i.e. QC pass or fail'
    ),
    (
        'note',
        'Pangolin QC Note',
        'Issues reported with Pangolin lineage assignment such as too many Ns in the input sequence'
        ' (Pangolin will not call a lineage for sequences with over 50% (0.5) N-content) or '
        'if the sequence is too short (e.g. "seq_len:0")'
    ),
]


def find_pangolin_lineage_csv(basedir: Path) -> Optional[Path]:
    for p in basedir.rglob(PANGOLIN_CSV):
        return p


def read_pangolin_csv(path: Path, sample_name: str = None) -> pd.DataFrame:
    df = pd.read_csv(path)
    df.sort_values('taxon', inplace=True)
    df.rename(columns={x: y for x, y, _ in pangolin_cols}, inplace=True)
    if sample_name:
        df['Sample'] = sample_name
    df.set_index('Sample', inplace=True)
    return df


def get_info(
        basedir: Path,
        pangolin_lineage_csv: Optional[Path] = None
) -> Optional[pd.DataFrame]:
    if pangolin_lineage_csv:
        return read_pangolin_csv(pangolin_lineage_csv)
    else:
        path = find_pangolin_lineage_csv(basedir)
        if path:
            return read_pangolin_csv(path)
        else:
            logger.info(f'Could not find single Pangolin output CSV with filename "{PANGOLIN_CSV}". '
                        f'Searching for Pangolin output per sample with sample name in filename')
            pangolin_outputs = find_file_for_each_sample(basedir=basedir,
                                                         glob_patterns=PANGOLIN_GLOB_PATTERNS,
                                                         sample_name_cleanup=PANGOLIN_SAMPLE_NAME_CLEANUP)
            if pangolin_outputs:
                return pd.concat([read_pangolin_csv(p, s) for s, p in pangolin_outputs.items()])
            else:
                return None



import re
from operator import itemgetter
from pathlib import Path
from typing import Dict, List, Tuple, Optional

from pydantic import BaseModel

from xlavir.util import find_file_for_each_sample

GLOB_PATTERNS = ['**/*.flagstat']


class SamtoolsFlagstat(BaseModel):
    sample: str
    n_total_reads: int
    n_mapped_reads: int


def parse_samtools_flagstat(p: Path) -> Tuple[int, int]:
    """Parse total and mapped number of reads from Samtools flagstat file"""
    total = 0
    mapped = 0
    with open(p) as fh:
        for line in fh:
            m = re.match(r'(\d+)', line)
            if m:
                if 'in total' in line:
                    total = int(m.group(1))
                if ' mapped (' in line:
                    mapped = int(m.group(1))
    return total, mapped


def select_flagstat(paths: List[Path]) -> Optional[Path]:
    xs = [(p, parse_samtools_flagstat(p)) for p in paths]
    xs.sort(key=itemgetter(1), reverse=True)
    try:
        return xs[0][0]
    except KeyError:
        return None


def get_info(basedir: Path) -> Dict[str, SamtoolsFlagstat]:
    out = {}
    flagstats = find_file_for_each_sample(basedir,
                                          glob_patterns=GLOB_PATTERNS,
                                          single_entry_selector_func=select_flagstat)
    # if multiple flagstat files are present, get stats from the one with the largest total number of reads
    for sample, flagstat_path in flagstats.items():
        n_total_reads, n_mapped_reads = parse_samtools_flagstat(flagstat_path)
        samtools_flagstat = SamtoolsFlagstat(sample=sample,
                                             n_mapped_reads=n_mapped_reads,
                                             n_total_reads=n_total_reads)
        out[sample] = samtools_flagstat
    return out



#!/usr/bin/env python

"""Tests for `xlavir.tools.nextflow` package."""

from pathlib import Path
from xlavir.tools.nextflow import exec_report

dirpath = Path(__file__).parent
input_html = dirpath / 'data/tools/nextflow/execution_report.html'

expected_report_info = {
    'workflow': 'nf-virontus',
    'execution_id': 'prickly_cuvier',
    'start_time': '17-Feb-2021 16:18:59',
    'completion_time': '17-Feb-2021 16:24:34',
    'command': 'nextflow run ../main.nf \\\n  --input samplesheet.csv \\\n  --genome false \\\n  --scov2 true \\\n  '
               '--freed \\\n  -resume \\\n  -profile docker \\\n  --tree_extra_fasta sequences.fasta \\\n  --tree '
               '\\\n  --outdir results',
    'project_directory': '/home/test/.nextflow/assets/peterk87/nf-virontus',
    'launch_directory': '/home/test/test',
    'workflow_profile': 'docker',
    'container': 'docker - peterk87/nf-virontus:2.0.0',
    'nextflow_version': 'version 20.10.0, build 5431 (01-11-2020 15:28 UTC)'
}


def test_parse_nextflow_exec_report_html():
    """Test finding and parsing of Nextflow workflow execution HTML report"""
    path = exec_report.find_exec_report(dirpath)
    assert str(path.absolute()) == str(input_html.absolute())
    info = exec_report.get_info(dirpath)
    assert isinstance(info, exec_report.NextflowWorkflowExecInfo)
    assert info.dict() == expected_report_info



import logging
from typing import Dict, List, Tuple

import pandas as pd

from xlavir.qc.quality_requirements import QualityRequirements
from xlavir.tools import mosdepth, samtools

logger = logging.getLogger(__name__)


def columns(low_coverage_threshold: int = 5, has_ct = False) -> List[Tuple[str, str, str]]:
    cols = [
        ('sample', 'Sample', 'Sample name'),
        ('ct_value', 'Ct Value', 'Real-time PCR Ct value') if has_ct else None,
        (
            'qc_status',
            'QC Status',
            'Quality control status, i.e. PASS or FAIL based on QC criteria '
            'such as minimum mean read depth '
            'or percent of reference genome covered by sequencing'
        ),
        (
            'qc_comment',
            'QC Comment',
            'Comments on any potential quality issues, i.e. why a sample did not pass QC.'
        ),
        (
            'genome_coverage',
            '% Genome Coverage',
            'Percent of reference genome sequence covered by sequencing'
        ),
        (
            'mean_coverage',
            'Mean Coverage Depth',
            'Mean sequencing coverage depth across entire reference genome sequence.'
        ),
        (
            'median_coverage',
            'Median Coverage Depth',
            'Median sequencing coverage depth across entire reference genome sequence.',
        ),
        (
            'n_total_reads',
            '# Total Reads',
            'Total number of raw reads obtained from sequencing for this sample.'
        ),
        (
            'n_mapped_reads',
            '# Mapped Reads',
            'Number of sequencing reads that mapped to the reference genome sequence.'
        ),
        (
            'n_zero_coverage',
            '# 0X positions',
            'Number of reference sequence positions with no coverage depth (0X), i.e. '
            'reference positions that did not have any reads spanning those positions.'
        ),
        (
            'n_low_coverage',
            f'# <{low_coverage_threshold}X positions',
            f'Number of reference sequence positions with fewer than {low_coverage_threshold} reads'
            f' spanning those positions.'
        ),
        (
            'ref_seq_length',
            'Reference Sequence Length',
            'Reference sequence length'
        ),
        (
            'zero_coverage_coords',
            '0X Coverage Regions',
            'A list of reference sequence 1-based regions with no coverage (0X).'
        ),
        (
            'low_coverage_coords',
            f'<{low_coverage_threshold}X Coverage Regions',
            f'A list of reference sequence 1-based regions with less than {low_coverage_threshold}'
            f' coverage depth.'
        ),
    ]
    return [x for x in cols if x is not None]


def report_format(df: pd.DataFrame, low_coverage_threshold: int = 5) -> pd.DataFrame:
    output_cols = columns(low_coverage_threshold)
    df.rename(columns={x: y for x, y, _ in output_cols}, inplace=True)
    df.set_index('Sample', inplace=True)
    return df


def create_qc_stats_dataframe(sample_depth_info: Dict[str, mosdepth.MosdepthDepthInfo],
                              sample_mapping_info: Dict[str, samtools.SamtoolsFlagstat],
                              sample_cts: Dict[str, float],
                              quality_reqs: QualityRequirements):
    sample_names = set(sample_depth_info.keys()) | set(sample_mapping_info.keys())
    logger.info(f'N samples: {len(sample_names)}')
    merged_stats_info = {}
    for sample in sample_names:
        depth_info = sample_depth_info[sample].dict() if sample in sample_depth_info else {}
        mapping_info = sample_mapping_info[sample].dict() if sample in sample_mapping_info else {}
        merged_stats_info[sample] = {**depth_info, **mapping_info}
        sample_ct = sample_cts.get(sample)
        if sample_ct is not None:
            merged_stats_info[sample]['ct_value'] = sample_ct
    df_stats = pd.DataFrame(list(merged_stats_info.values()))
    mask_pass_depth = (df_stats.median_coverage >= quality_reqs.min_median_depth)
    mask_pass_breadth = (df_stats.genome_coverage >= quality_reqs.min_genome_coverage)
    qc_pass_mask = mask_pass_depth & mask_pass_breadth
    df_stats['qc_status'] = 'FAIL'
    df_stats.loc[qc_pass_mask, 'qc_status'] = 'PASS'
    qc_comments = []
    for i, (pass_depth, pass_breadth) in enumerate(zip(mask_pass_depth, mask_pass_breadth)):
        comments = []
        if not pass_depth:
            comments += [f'Median depth below {quality_reqs.min_median_depth}']
        if not pass_breadth:
            comments += [f'Genome coverage below {quality_reqs.min_genome_coverage:.0%}']
        qc_comments.append('; '.join(comments))
    df_stats['qc_comment'] = qc_comments
    df_stats.loc[qc_pass_mask, 'qc_comment'] = ''
    df_stats.sort_values('sample', inplace=True)

    present_cols = set(df_stats.columns)
    output_cols = columns(
        quality_reqs.low_coverage_threshold,
        has_ct=bool(sample_cts)
    )
    df_stats = df_stats.loc[:, [x for x, y, _ in output_cols if x in present_cols]]

    logger.debug(df_stats)
    return df_stats



"""Main module."""
import logging
from pathlib import Path
from typing import Optional, List

from xlavir import qc
from xlavir.io import ct
from xlavir.io.excel_sheet_dataframe import ExcelSheetDataFrame, SheetName
from xlavir.tools import mosdepth, samtools, consensus, pangolin, variants, nextclade, fastp
from xlavir.tools.nextflow import exec_report
from xlavir.tools.nextflow.exec_report import to_dataframe

logger = logging.getLogger(__name__)


def run(
        input_dir: Path,
        quality_reqs: Optional[qc.QualityRequirements],
        pangolin_lineage_csv: Optional[Path] = None,
        ct_values_table: Optional[Path] = None
) -> List[ExcelSheetDataFrame]:
    if quality_reqs is None:
        quality_reqs = qc.QualityRequirements()
    nf_exec_info = exec_report.get_info(input_dir)
    sample_depth_info = mosdepth.get_info(input_dir, low_coverage_threshold=quality_reqs.low_coverage_threshold)
    if logger.level == logging.DEBUG:
        for sample, info in sample_depth_info.items():
            logger.debug(info.dict())
    sample_mapping_info = samtools.get_info(input_dir)
    if logger.level == logging.DEBUG:
        for sample, info in sample_mapping_info.items():
            logger.debug(info.dict())
    sample_total_reads = fastp.get_info(input_dir)
    for sample, total_reads in sample_total_reads.items():
        mapping_info = sample_mapping_info.get(sample, None)
        if mapping_info is None:
            logger.warning(
                f'Found fastp total reads ({total_reads}) for sample "{sample}", '
                f'but no such sample exists in sample mapping info dict: '
                f'{list(sample_mapping_info.keys())}'
            )
            continue
        if total_reads != mapping_info.n_total_reads:
            logger.info(
                f'Found fastp total reads ({total_reads}) for sample "{sample}". '
                f'Samtools mapping info indicates total reads as '
                f'{mapping_info.n_total_reads} '
                f'({mapping_info.n_mapped_reads} mapped). '
                f'Using fastp value for total reads.'
            )
            mapping_info.n_total_reads = total_reads
    sample_cts = ct.read_ct_table(ct_values_table) if ct_values_table else {}
    sample_variants = variants.get_info(input_dir, qc_reqs=quality_reqs)

    dfs: List[ExcelSheetDataFrame] = []
    df_stats = qc.create_qc_stats_dataframe(sample_depth_info,
                                            sample_mapping_info,
                                            sample_cts=sample_cts,
                                            quality_reqs=quality_reqs)
    dfs.append(ExcelSheetDataFrame(sheet_name=SheetName.qc_stats.value,
                                   df=qc.report_format(df_stats,
                                                       low_coverage_threshold=quality_reqs.low_coverage_threshold),
                                   pd_to_excel_kwargs=dict(freeze_panes=(1, 1), na_rep='NA'),
                                   header_comments={x: y for _, x, y in
                                                    qc.columns(quality_reqs.low_coverage_threshold)}))
    df_pangolin = pangolin.get_info(basedir=input_dir,
                                    pangolin_lineage_csv=pangolin_lineage_csv)
    if df_pangolin is not None:
        dfs.append(ExcelSheetDataFrame(sheet_name=SheetName.pangolin.value,
                                       df=df_pangolin,
                                       pd_to_excel_kwargs=dict(freeze_panes=(1, 1)),
                                       header_comments={x: y for _, x, y in pangolin.pangolin_cols}))
    sample_nextclade = nextclade.get_info(basedir=input_dir)
    if sample_nextclade:
        dfs.append(ExcelSheetDataFrame(sheet_name=SheetName.nextclade.value,
                                       df=nextclade.to_dataframe(sample_nextclade),
                                       pd_to_excel_kwargs=dict(freeze_panes=(1, 1)),
                                       header_comments={x: y for _, x, y in nextclade.nextclade_cols}))
    if sample_variants:
        df_variants = variants.to_dataframe(sample_variants.values())
        dfs.append(ExcelSheetDataFrame(sheet_name=SheetName.variants.value,
                                       df=df_variants,
                                       pd_to_excel_kwargs=dict(freeze_panes=(1, 1)),
                                       include_header_width=False,
                                       header_comments={name: desc for _, name, desc in variants.variants_cols}))
        if 'Mutation' in df_variants.columns:
            df_varsum = variants.to_summary(df_variants)
            dfs.append(ExcelSheetDataFrame(sheet_name=SheetName.varsum.value,
                                           df=df_varsum,
                                           pd_to_excel_kwargs=dict(freeze_panes=(1, 1)),
                                           header_comments={name: desc for _, name, desc in
                                                            variants.variant_summary_cols + variants.variants_cols}))
            df_varmap = variants.to_variant_pivot_table(df_variants)
            max_index_length = df_varmap.index.str.len().max()
            dfs.append(ExcelSheetDataFrame(sheet_name=SheetName.varmat.value,
                                           df=df_varmap,
                                           pd_to_excel_kwargs=dict(freeze_panes=(1, 1), na_rep=0.0),
                                           autofit=False,
                                           column_widths=[max_index_length + 2] + [3 for _ in
                                                                                   range(df_varmap.columns.size)]))
        else:
            logger.warning(
                'No column "Mutation" found in variant info dataframe. '
                'SnpEff/SnpSift table may not have been found or parsed correctly.'
            )

    dfs.append(ExcelSheetDataFrame(sheet_name=SheetName.consensus.value,
                                   df=consensus.get_info(basedir=input_dir),
                                   autofit=False,
                                   pd_to_excel_kwargs=dict(index=None, header=None)))
    if nf_exec_info:
        df_exec_info = to_dataframe(nf_exec_info)
        dfs.append(ExcelSheetDataFrame(sheet_name=SheetName.workflow_info.value,
                                       df=df_exec_info,
                                       autofit=False,
                                       column_widths=(20, 100)))
        logger.debug(df_exec_info)
    else:
        logger.warning(f'Could not find "execution_report.html" in "{input_dir}"!')

    return dfs



"""Console script for xlavir."""
import logging
from enum import Enum
from pathlib import Path
from typing import Optional, List
from sys import version_info

import typer
from rich.logging import RichHandler
import pandas as pd

from xlavir.__about__ import __version__
from xlavir.images import get_images_for_sheets
from xlavir.io.excel_sheet_dataframe import ExcelSheetDataFrame, SheetName
from xlavir.qc import QualityRequirements
from xlavir.xlavir import run
from xlavir.io.xl import write_xlsx_report

app = typer.Typer()


class QCPresets(str, Enum):
    default = 'default'
    scov2_illumina = 'scov2_illumina'
    scov2_nanopore = 'scov2_nanopore'


qc_presets = dict(
    default=QualityRequirements(),
    scov2_illumina=QualityRequirements(
        min_genome_coverage=0.95,
        min_median_depth=30,
        low_coverage_threshold=5,
    ),
    scov2_nanopore=QualityRequirements(
        min_genome_coverage=0.95,
        min_median_depth=50,
        low_coverage_threshold=10,
    )
)


def version_callback(value: bool):
    if value:
        typer.echo(f"xlavir version {__version__}")
        raise typer.Exit()


def check_dir_exists_callback(path: Path) -> Path:
    if not (path.exists() or path.is_dir()):
        raise typer.BadParameter(f'An existing Nextflow workflow results directory must be specified! '
                                 f'"{path}" does not exist or is not a directory!')
    return path


@app.command(
    epilog=f'xlavir version {__version__}; Python {version_info.major}.{version_info.minor}.{version_info.micro}')
def main(
        input_dir: Path = typer.Argument(..., callback=check_dir_exists_callback),
        output: Path = typer.Argument('xlavir-report.xlsx'),
        ct_table: Path = typer.Option(None, help='Table of sample IDs and rtPCR Ct values'),
        pangolin_lineage_csv: Path = typer.Option(None, help='Pangolin lineage report CSV'),
        qc_preset: Optional[QCPresets] = typer.Option(None, help='Quality check preset'),
        low_coverage_threshold: Optional[int] = typer.Option(None, help='Low coverage threshold. '
                                                                        'Used for calculation of % genome coverage.'),
        min_genome_coverage: Optional[float] = typer.Option(None, help='Min genome coverage. e.g. 0.95 == 95%'),
        min_median_depth: Optional[int] = typer.Option(None, help='Min median coverage depth'),
        major_allele_freq: float = typer.Option(0.75, help='Major alternate allele fraction'),
        spreadsheet: Optional[List[Path]] = typer.Option(None, help='Copy Excel worksheet from workbook. '
                                                                    'Can specify multiple.'),
        image: Optional[List[Path]] = typer.Option(None, help="Image path for image to add to sheet. "
                                                              "Can specify multiple."),
        image_title: Optional[List[str]] = typer.Option(None, help="Image sheet title"),
        image_description: Optional[List[str]] = typer.Option(None, help="Image description."),
        verbose: bool = typer.Option(default=False, help='Verbose logging'),
        version: Optional[bool] = typer.Option(None, callback=version_callback,
                                               help=f'Print "xlavir version {__version__}" and exit'),
):
    """xlavir - create an Excel report from a bioinformatics analysis output directory

    Typical usage:

    $ xlavir /path/to/viralrecon-or-virontus-results

    Will output "report.xlsx" in current directory.

    Specify report output filename:

    $ xlavir /path/to/viralrecon-or-virontus-results xlavir-run-XXXX.xlsx

    """
    from rich.traceback import install
    install(show_locals=True, width=240, word_wrap=True)

    logging.basicConfig(
        format='%(message)s',
        datefmt='[%Y-%m-%d %X]',
        level=logging.DEBUG if verbose else logging.INFO,
        handlers=[
            RichHandler(rich_tracebacks=True, tracebacks_show_locals=True)
        ],
    )

    images_for_sheets = None
    if image:
        images_for_sheets = get_images_for_sheets(image, image_title, image_description)

    quality_reqs = qc_presets[QCPresets.default.value]
    if qc_preset:
        quality_reqs = qc_presets[qc_preset.value]
    if min_genome_coverage:
        quality_reqs.min_genome_coverage = min_genome_coverage
    if low_coverage_threshold:
        quality_reqs.low_coverage_threshold = low_coverage_threshold
    if min_median_depth:
        quality_reqs.min_median_depth = min_median_depth
    if major_allele_freq:
        quality_reqs.major_allele_freq = major_allele_freq

    dfs = run(input_dir=input_dir,
              pangolin_lineage_csv=pangolin_lineage_csv,
              ct_values_table=ct_table,
              quality_reqs=quality_reqs)
    dfs.append(ExcelSheetDataFrame(
        sheet_name=SheetName.xlavir_info.value,
        df=pd.DataFrame([
            ('xlavir version', __version__),
            ('Python version', f'{version_info.major}.{version_info.minor}.{version_info.micro}'),
            ('Input directory', input_dir.absolute()),
            ('QC Requirements', quality_reqs),
        ], columns=['Attribute', 'Value']).set_index('Attribute')
    ))
    write_xlsx_report(dfs=dfs,
                      output_xlsx=output,
                      quality_reqs=quality_reqs,
                      images_for_sheets=images_for_sheets)
    if spreadsheet:
        from xlavir.io.xl import copy_spreadsheet
        for excel in spreadsheet:
            copy_spreadsheet(excel, output)

    return 0


if __name__ == "__main__":
    app()  # pragma: no cover



import re
from typing import Optional

import pandas as pd
from bs4 import BeautifulSoup
import logging
from pydantic import BaseModel
from pathlib import Path

logger = logging.getLogger(__name__)


class NextflowWorkflowExecInfo(BaseModel):
    """Nextflow workflow execution information"""
    workflow: str
    execution_id: str
    start_time: str
    completion_time: str
    command: str
    project_directory: str
    launch_directory: str
    workflow_profile: str
    container: Optional[str] = None
    nextflow_version: str


def find_exec_report(basedir: Path) -> Path:
    """Find most recently modified Nextflow execution report in a directory

    It is assumed that there will be at least one Nextflow execution report
    in the directory if it is a Nextflow output directory.

    Args:
        basedir (Path): Directory to search for Nextflow execution reports

    Returns:
        Path: Path to most recently modified Nextflow execution report

    Raises:
        FileNotFoundError: If no Nextflow execution report is found in the directory
    """
    reports = list(basedir.rglob('execution_report*.html'))
    if reports:
        reports.sort(key=lambda x: x.stat().st_mtime, reverse=True)
        return reports[0]
    raise FileNotFoundError(f'No Nextflow execution report found in "{basedir}"')


def parse(path: Path) -> NextflowWorkflowExecInfo:
    """Parse Nextflow execution HTML report

    Args:
        path (Path): Path to Nextflow execution report

    Returns:
        NextflowWorkflowExecInfo: Parsed Nextflow execution report
    """
    with open(path) as fh:
        soup = BeautifulSoup(fh, 'html.parser')
        execution_id = re.sub(r'\[(\w+)].*', r'\1', soup.find('head').find('title').text)
        command = soup.find(attrs={'class': 'nfcommand'}).text
        workflow_start = soup.find(id='workflow_start').text
        workflow_complete = soup.find(id='workflow_complete').text
        wf_attrs = soup.find(class_='container').find(class_='row').text.strip().split('\n')
        wf_attrs = {wf_attrs[i]: wf_attrs[i + 1] for i in range(0, len(wf_attrs), 2)}
        project_dir = wf_attrs.get('Project directory', 'UNKNOWN')
        if project_dir:
            workflow = Path(project_dir).name
        container_engine = wf_attrs.get('Container engine', None)
        wf_container = wf_attrs.get('Workflow container', None)
        container = None
        if wf_container and container_engine:
            container = f'{container_engine} - {wf_container}'
        nextflow_version = wf_attrs.get('Nextflow version', 'UNKNOWN')
        launch_directory = wf_attrs.get('Launch directory', 'UNKNOWN')
        workflow_profile = wf_attrs.get('Workflow profile', 'NA')

        return NextflowWorkflowExecInfo(workflow=workflow,
                                        execution_id=execution_id,
                                        start_time=workflow_start,
                                        completion_time=workflow_complete,
                                        command=command.replace(' -', ' \\\n  -'),
                                        container=container,
                                        nextflow_version=nextflow_version,
                                        project_directory=project_dir,
                                        launch_directory=launch_directory,
                                        workflow_profile=workflow_profile)


def get_info(basedir: Path) -> Optional[NextflowWorkflowExecInfo]:
    exec_report_path = find_exec_report(basedir)
    if exec_report_path:
        logger.info(f'Found Nextflow execution report "{exec_report_path}"')
        return parse(exec_report_path)
    else:
        logger.warning(f'Could not find Nextflow execution report in "{basedir}". '
                       f'Did you specify the input directory as a Nextflow output directory?')
        return None


def to_dataframe(nf_exec_info):
    df_exec_info = pd.DataFrame([(x, y) for x, y in nf_exec_info.dict().items()])
    df_exec_info.columns = ['Attribute', 'Value']
    df_exec_info.set_index('Attribute', inplace=True)
    return df_exec_info
